"""Functions for reading DICOM files ith dicomity"""

from os.path import join
from typing import Optional

import numpy as np
from pydicom import filereader, FileDataset

from dicomity.dictionary import DicomDictionary
from dicomity.types import GroupingMetadata


def read_dicom_image(file_name: str) -> np.array:
    meta = filereader.dcmread(file_name, stop_before_pixels=False)
    return meta.pixel_array


def read_dicom_tags(
        file_name: str,
        dictionary: Optional[DicomDictionary] = None,
        specific_tags: Optional[list[str]] = None
) -> FileDataset:
    if dictionary:
        specific_tags = dictionary.allTags()
    meta = filereader.dcmread(
        file_name,
        stop_before_pixels=True,
        specific_tags=specific_tags
    )
    return meta


def read_grouping_metadata(file_name: str) -> GroupingMetadata:
    meta = filereader.dcmread(fp=file_name,
                              stop_before_pixels=True,
                              specific_tags=GroupingMetadata.groupingTags())
    return GroupingMetadata.fromPyDicom(meta)


def is_dicom(file_name: str) -> bool:
    """Test whether a file is in DICOM format

    Uses a simple test for the DICM preamble - this is very fast but
    may occasionally produce false positives.

    For a more robust guarantee that a file really is DICOM, first call this
    function and if True, try to parse the file - if the tag data cannot be
    parsed then it is likely not a DICOM file

    Args:
        file_name: path and filename of the file to test

    Returns:
        True if the file appears to be a Dicom file
    """

    with open(file_name, "rb") as f:
        f.seek(128)
        preamble = f.read(4)
        return preamble == b"DICM"


def is_dicom_image_file(file_path: str, file_name: str):
    """Tests if a file is a DICOM file but not a DICOMDIR file

    Args:
        file_path:
        file_name:

    Returns:
        True if the file is DICOM and not a DICOMDIR
    """
    if file_name == 'DICOMDIR':
        return False

    return is_dicom(join(file_path, file_name))



from os.path import join
from pathlib import Path

from dicomity.util import current_date_string, get_user_directory
from pyreporting.progress import TerminalProgress, ProgressStackItem
from pyreporting.util import get_calling_function, throw_exception


class Reporting:
    """Provides an interface for error and progress reporting.

    Reporting is the interface used to process errors, warnings, messages,
    logging information and progress reports. This means that warnings,
    errors and progress are handled via a callback instead of
    directly bringing up message and progress boxes or writing to the
    command window. This allows applications to choose how they
    process error, warning and progress information.

    You can create your own implementation of this interface to get
    customised message behaviour; for example, if you are running a batch
    script you may wish to write all messages to a log file instead of
    displaying them on the command line.

    This is an abstract class; you should not directly create an instance
    of Reporting. Instead, you should either use one of the
    existing implementation classes (CoreReporting, CoreReportingDefault) or
    you can create your own to achieve customised behaviour.
    """

    def log(self, identifier, message):
        """Write debugging information to the log file"""
        raise NotImplementedError

    def logVerbose(self, identifier, message):
        """Write debugging information to the log file, but only if in verbose
        mode"""
        raise NotImplementedError

    # Displays an information message to the user. This will generally be
    # written to the command window
    def showMessage(self, identifier, message, supplementary_info=None):
        raise NotImplementedError

    # Displays a warning to the user. This will generally be
    # written to the command window.
    def showWarning(self, identifier, message, supplementary_info=None):
        raise NotImplementedError

    # Displays an error to the user. This may trigger an
    # exception which will ultimately be displayed to the user as a modal
    # error dialog or written to the command window,
    def error(self, identifier, message):
        raise NotImplementedError

    # Displays an error to the user. Similar to Error() but also displays
    # additional information about an exception which triggered the error.
    def errorFromException(self, identifier, message, ex):
        raise NotImplementedError

    # Displays the progress dialog with the specified title
    def show_progress(self, text):
        raise NotImplementedError

    # Hides the progress dialog
    def completeProgress(self):
        raise NotImplementedError

    # Changes the subtext in the progress dialog
    def updateProgressMessage(self, text):
        raise NotImplementedError

    # Changes the percentage complete in the progress dialog, displaying if
    # necessary
    def updateProgressValue(self, progress_value):
        raise NotImplementedError

    # When progress reporting consists of a number of stages, use this
    # method to ensure progress is handled correctly
    def updateProgressStage(self, progress_stage, num_stages):
        raise NotImplementedError

    # Changes the percentage complete and subtext in the progress dialog,
    # displaying if necessary
    def updateProgressAndMessage(self, progress_value, text):
        raise NotImplementedError

    def hasBeenCancelled(self):
        """Used to check if the user has clicked the cancel button in the
        progress dialog"""
        raise NotImplementedError

    def checkForCancel(self):
        """Force an exception if the user has clicked cancel in the progress
        dialog"""
        raise NotImplementedError

    def pushProgress(self):
        """Nests progress reporting. After calling this function, subsequent
        progress updates will modify the progress bar between the current
        value ane the current value plus the last value_change."""
        raise NotImplementedError

    def popProgress(self):
        """Remove one layer of progress nesting, returning to the previous
        progress reporting."""
        raise NotImplementedError

    def clearProgressStack(self):
        """Clear all progress nesting"""
        raise NotImplementedError

    def showAndClearPendingMessages(self):
        """Show any error or warning messages and clear the message stack"""
        raise NotImplementedError


class ReportingBase(Reporting):
    """Provides error, message and progress reporting.

    Implementation of Reporting, which is used by
    CoreMat and other libraries for progress and error/message reporting. This
    implementation displays warnings and messages on the command window,
    and uses Matlab's error() command to process errors. Logging
    information is writen to a log file. An object which implements
    the Progress can be passed in for progress reporting.

    Usage
    -----

    You should create a single CoreReporting object and pass it into all the
    CoreMat routines you use in order to provide error, warning,
    message and progress reporting during execution of routines.

    If you are not writing a gui application but would like a standard
    pop-up progress dialog to appear while waiting for plugins to execute,
    consider creating a CoreReportingDefault object instead. Use CoreReporting
    if you want to specify your own progress dialog, or specify a gui
    viewing panel, or if you want no progress dialog at all.

        reporting = CoreReporting(progress_dialog, viewing_panel);

            progress_dialog - a Progress object such as
                a TerminalProgress for displaying a progress bar. You can
                omit this argument
                or replace it with None if you are writing scripts to run
                in the background and do not want progress dialogs popping
                up. Otherwise, you should create a CoreProgressDialog
                or else implement your own progress class if you want
                custom behaviour.

    See reporting.py for details of the methods this class implements.
    """

    CancelErrorId = 'CoreReporting:UserCancel'

    def __init__(self, progress_dialog=None,
                 verbose_mode=False,
                 log_file_name=None):

        # Handle to a ProgressInterface object
        self.ProgressDialog = progress_dialog

        self.VerboseMode = verbose_mode
        Path(get_user_directory()).mkdir(parents=True,
                                                     exist_ok=True)
        if not log_file_name:
            log_file_name = join(get_user_directory(),
                                 'reporting.log')
        self.LogFileName = log_file_name

        # self.ProgressStack = []
        # self.CurrentProgressStackItem = CoreProgressStackItem('', 0, 100)
        # self.ParentProgressStackItem = CoreProgressStackItem('', 0, 100)
        self.clearProgressStack()

    # properties (Access = private)
    #     LogFileName     # Full path to log file
    #
    #     # Stack for nested progress reporting
    #     ProgressStack
    #     CurrentProgressStackItem
    #     ParentProgressStackItem
    #     VerboseMode
    # end
    #

    def log(self, identifier, message):
        calling_function = get_calling_function(2)
        self._appendToLogFile(f'{calling_function}: {identifier}: {message}')

    def log_verbose(self, identifier, message):
        if self.VerboseMode:
            calling_function = get_calling_function(2)
            self._appendToLogFile(
                f'{calling_function}: {identifier}: {message}')

    def show_message(self, identifier, message, supplementary_info=None):
        calling_function = get_calling_function(2)
        if not calling_function:
            calling_function = 'Command Window'
        print(message)

        self._appendToLogFile(f'{calling_function}: {identifier}: {message}')

        if supplementary_info:
            print('Additional information on this message:')
            print(supplementary_info)

    def show_message_from_exception(self, identifier, message, ex):
        calling_function = get_calling_function(2)
        if not calling_function:
            calling_function = 'Command Window'
        message = f'{calling_function}: {identifier}: {message}; ' \
                  f'Exception message:{ex.message} : {ex.stack(1).name}'
        summary, stack_text = self._getExceptionSummary(ex)

        self._appendToLogFile(message)
        self._appendToLogFile(summary)
        self._appendToLogFile(stack_text)
        print(message)
        print(' ')
        print('If reporting an error, please include the following '
              'information:')
        print(f'  {summary}')
        print(f'  {stack_text} \n')
        print(' ')

    def show_warning(self, identifier, message, supplementary_info=None):
        calling_function = get_calling_function(2)

        self._appendToLogFile(f'{calling_function}: '
                              f'WARNING: {identifier}:{message}')
        print(f'WARNING: {message}')
        if supplementary_info:
            print('Additional information on this warning:')
            print(supplementary_info)

    def error(self, identifier, message):
        calling_function = get_calling_function(2)

        message_text = f'Error in function {calling_function}: {message}'
        self._appendToLogFile(f'{calling_function}: '
                              f'ERROR: {identifier}:{message}')

        throw_exception(message=message_text, identifier=identifier)

    def error_from_exception(self, identifier, message, ex):
        calling_function = get_calling_function(2)

        summary, stack_text = self._getExceptionSummary(ex)
        message_text = f'Error in function {calling_function}: {identifier}: ' \
                       f'{message}; Exception message: {str(ex)}'

        self._appendToLogFile(message_text)
        self._appendToLogFile(summary)
        self._appendToLogFile(stack_text)

        print(message_text)
        print(' ')
        print('If reporting an error, please include the following information')
        print('  ' + summary)
        print('  ' + stack_text)
        print(' ')

        throw_exception(message=message_text, identifier=identifier,
                        exception=ex)

    def show_progress(self, text=None):
        adjusted_text = self._adjustProgressText(text)

        if self.ProgressDialog:
            self.ProgressDialog.set_progress_text(adjusted_text)
            self.CurrentProgressStackItem.visible = True

    def complete_progress(self):
        if self.ProgressDialog:
            if not self.ProgressStack or \
                    not self.ParentProgressStackItem.visible:
                self.ProgressDialog.complete()
                self.CurrentProgressStackItem.visible = False

    def update_progress_message(self, text):
        adjusted_text = self._adjustProgressText(text)

        if self.ProgressDialog:
            self.ProgressDialog.set_progress_text(adjusted_text)
            self.CurrentProgressStackItem.visible = True

    def update_progress_value(self, progress_value):
        adjusted_value = self._adjustProgressValue(progress_value, None)

        if self.ProgressDialog:
            self.ProgressDialog.set_progress_value(adjusted_value)
            self.CurrentProgressStackItem.visible = True
        self.checkForCancel()

    def updateProgressStage(self, progress_stage, num_stages):
        progress_value = 100*progress_stage/num_stages
        value_change = 100/num_stages
        adjusted_value = self._adjustProgressValue(progress_value, value_change)
        if self.ProgressDialog:
            self.ProgressDialog.SetProgressValue(adjusted_value)
            self.CurrentProgressStackItem.visible = True
        self.checkForCancel()

    def updateProgressAndMessage(self, progress_value, text):
        adjusted_value = self._adjustProgressValue(progress_value, None)
        adjusted_text = self._adjustProgressText(text)

        if self.ProgressDialog:
            self.ProgressDialog.set_progress_and_message(adjusted_value,
                                                         adjusted_text)
            self.CurrentProgressStackItem.visible = True

        self.checkForCancel()

    def hasBeenCancelled(self):
        if self.ProgressDialog:
            return self.ProgressDialog.cancel_clicked()
        else:
            return False

    def checkForCancel(self):
        if self.hasBeenCancelled():
            self.error(ReportingBase.CancelErrorId, 'User cancelled')

    def push_progress(self):
        self.ProgressStack.append(self.ParentProgressStackItem)
        self.ParentProgressStackItem = self.CurrentProgressStackItem
        self.CurrentProgressStackItem = ProgressStackItem(
            progress_text='',
            min_position=self.ParentProgressStackItem.min_position,
            max_position=self.ParentProgressStackItem.max_position,
            visible=self.ParentProgressStackItem.visible
        )

    def popProgress(self):
        self.CurrentProgressStackItem = self.ParentProgressStackItem
        if not self.ProgressStack:
            self.ParentProgressStackItem = None
        else:
            self.ParentProgressStackItem = self.ProgressStack.pop()

    def clearProgressStack(self):
        self.ProgressStack = []
        self.CurrentProgressStackItem = ProgressStackItem('', 0, 100)
        self.ParentProgressStackItem = ProgressStackItem('', 0, 100)

    def showAndClearPendingMessages(self):
        pass

    def openPath(self, file_path, message):
        print(f'{message}: {file_path}')

    def _appendToLogFile(self, message):
        message = current_date_string() + ': ' + message
        with open(self.LogFileName, 'a') as file:
            file.write(message)

    @staticmethod
    def _getExceptionSummary(exc):
        stack = 'Stack: TODO'
        # stack = 'Stack:' + traceback.format_exc(exc)
        summary = 'Summary: TODO'
        # summary = f'Error in {exc.stack(1).name} (CORE {CORE_VERSION}) : ' \
        #           f'{exc.message}'
        return summary, stack

    def _adjustProgressText(self, text):
        adjusted_text = text
        self.CurrentProgressStackItem.progress_text = text
        return adjusted_text

    def _adjustProgressValue(self, value, value_change=None):
        if not value_change:
            value_change = value - \
                           self.CurrentProgressStackItem.last_progress_value
        self.CurrentProgressStackItem.last_progress_value = value

        scale = (self.ParentProgressStackItem.max_position -
                 self.ParentProgressStackItem.min_position) / 100
        adjusted_value = self.ParentProgressStackItem.min_position + scale*value
        self.CurrentProgressStackItem.min_position = adjusted_value
        if value_change > 0:
            self.CurrentProgressStackItem.max_position = adjusted_value + \
                                                         scale * value_change
        return adjusted_value

    def _setValueChange(self, value_change):
        value = self.CurrentProgressStackItem.last_progress_value
        scale = (self.ParentProgressStackItem.max_position -
                 self.ParentProgressStackItem.min_position) / 100
        adjusted_value = self.ParentProgressStackItem.min_position + scale*value
        self.CurrentProgressStackItem.min_position = adjusted_value
        if value_change > 0:
            self.CurrentProgressStackItem.max_position = adjusted_value + \
                                                         scale * value_change


class ReportingDefault(ReportingBase):
    """Provides error, message and progress reporting.

    CoreReporting. Implementation of Reporting, which is used by
    CoreMat and related libraries for progress and error/message reporting. This
    is a convenient implementation with no constructor arguments which
    creates a progress dialog and writes messages to the command window,
    but which has no callbacks to the gui.

    This class is intended for use by CoreMat library functions
    when no reporting object is given. It can also be used in your code
    to create a default progress reporting implementation for input into
    CoreMat and related libraries.

    See reporting.py for details of the methods this class
    implements."""

    def __init__(self):
        super().__init__(TerminalProgress())



from dataclasses import dataclass

from pyreporting.reporting import Reporting, ReportingDefault


class ReportingWithCache(Reporting):
    """Provides error, message and progress reporting, with
    warnings and messages cached to prevent display of duplicates
    
    ReportingWithCache is a wrapper around a Reporting object.
    Messages and warnings are cached using a ReportingWarningsCache and
    displayed at the end of the algorith. Duplicate messages and warnings are
    grouped together to prevent multiple messages appearing.
    
    See reporting.py for details of the methods this class implements.
    """
    def __init__(self, reporting=None):
        if not reporting:
            reporting = ReportingDefault()
        self._reporting = reporting
        self._warnings_cache = CoreReportingWarningsCache(self)

    def delete(self):
        self.showAndClearPendingMessages()
    
    def clearProgressStack(self):
        self._reporting.clearProgressStack()
    
    def pushProgress(self):
        self._reporting.pushProgress()

    def popProgress(self):
        self._reporting.popProgress()

    def log(self, identifier, message):
        self._reporting.logVerbose(identifier=identifier, message=message)

    def logVerbose(self, identifier, message):
        self._reporting.logVerbose(identifier=identifier, message=message)

    def showMessage(self, identifier, message, supplementary_info=None):
        self._warnings_cache.addPendingMessages(
            identifier, message, supplementary_info)

    def showMessageFromException(self, identifier, message, ex):
        self._reporting.showMessageFromException(identifier, message, ex)

    def showWarning(self, identifier, message, supplementary_info=None):
        self._warnings_cache.addPendingWarning(
            identifier, message, supplementary_info)

    def error(self, identifier, message):
        self._reporting.error(identifier, message)

    def errorFromException(self, identifier, message, ex):
        self._reporting.errorFromException(identifier, message, ex)

    def show_progress(self, text):
        self._reporting.show_progress(text)

    def completeProgress(self):
        self._reporting.completeProgress()

    def updateProgressMessage(self, text):
        self._reporting.updateProgressMessage(text)

    def updateProgressValue(self, progress_value):
        self._reporting.updateProgressValue(progress_value)

    def updateProgressAndMessage(self, progress_value, text):
        self._reporting.updateProgressAndMessage(progress_value, text)

    def updateProgressStage(self, progress_stage, num_stages):
        self._reporting.updateProgressStage(progress_stage, num_stages)

    def hasBeenCancelled(self):
        return self._reporting.hasBeenCancelled()

    def checkForCancel(self):
        self._reporting.checkForCancel()

    def showAndClearPendingMessages(self):
        self._warnings_cache.showAndClear()
        self._reporting.showAndClearPendingMessages()

    def showCachedMessage(self, identifier, message, supplementary_info):
        self._reporting.showMessage(identifier, message)

    def showCachedWarning(self, identifier, message, supplementary_info):
        self._reporting.showWarning(identifier, message, supplementary_info)

    def openPath(self, file_path, message):
        self._reporting.openPath(file_path, message)


@dataclass
class PendingRecord:
    ID: str
    Text: str
    SupplementaryInfo: str
    Count: int = 1


class CoreReportingWarningsCache:
    def __init__(self, reporting_with_cache):
        self._reporting_with_cache = reporting_with_cache
        self._pending_messages = {}
        self._pending_warnings = {}

    def addPendingWarning(self, warning_id, warning_text, supplementary_info):
        if warning_id in self._pending_warnings:
            self._pending_warnings[warning_id].Count += 1
        else:
            self._pending_warnings[warning_id] = PendingRecord(
                ID=warning_id,
                Text=warning_text,
                SupplementaryInfo=supplementary_info
            )

    def addPendingMessages(self, message_id, message_text, supplementary_info):
        if message_id in self._pending_messages:
            self._pending_messages[message_id].Count += 1
        else:
            self._pending_warnings[message_id] = PendingRecord(
                ID=message_id,
                Text=message_text,
                SupplementaryInfo=supplementary_info
            )

    def showAndClear(self):
        for warning_id, warning in self._pending_warnings.items():
            warning_text = warning.Text
            if warning.Count > 1:
                warning_text = f'(repeated x{warning.Count}) {warning.Text}'
            self._reporting_with_cache.showCachedWarning(
                identifier=warning_id,
                message=warning_text,
                supplementary_info=warning.SupplementaryInfo)
        self._pending_warnings.clear()

        for message_id, message in self._pending_messages.items():
            message_text = message.Text
            if message.Count > 1:
                message_text = f'(repeated x{message.Count}) {message.Text}'
            self._reporting_with_cache.showCachedMessage(
                identifier=message_id,
                message=message_text,
                supplementary_info=message.SupplementaryInfo)
        self._pending_messages.clear()


from dataclasses import dataclass

from tqdm import tqdm

from dicomity.util import remove_html


class Progress:
    """Interface for classes which implement a progress bar.

    The Reporting class uses this interface to display and update a
    progress bar and associated text. The ProgressDialog class
    implements the progress bar as a standard Matlab progress dialog,
    whereas other implementations may implement a custom progress panel for an
    existing figure.

    You can implement custom progress reporting by creating a class which
    implements this interface, and then passing an instance of your
    progress class into Reporting.
    """

    def resize(self, panel_position):
        """Indicates the gui has been resized"""
        raise NotImplementedError
           
    def show_and_hold(self, text):
        """Show the progress bar, and keep it displayed until hide() called"""

    def hide(self):
        """Hide the progress bar"""
        raise NotImplementedError

    def complete(self):
        """Call to complete a progress operation, which will also hide the
        dialog unless the dialog has been held by show_and_hold()"""
        raise NotImplementedError

    def set_progress_text(self, text):
        """Change the subtext in the progress dialog"""
        raise NotImplementedError

    def set_progress_value(self, progress_value):
        """Changes the value of the progress bar"""
        raise NotImplementedError

    def set_progress_and_message(self, progress_value, text):
        """Change the value of the progress bar and the subtext"""
        raise NotImplementedError

    def cancel_clicked(self):
        """Check if the cancel button was clicked by the user"""
        raise NotImplementedError


class TerminalProgress(Progress):
    """A dialog used to report progress information

    TerminalProgress creates and manages a waitbar to mark progress
    in operations performed by libraries and applications which use
    pyreporting.
    It provides a default implementation of
    Progress that can be used by your own code or can be
    instantiated automatically be functions when no progress
    interface object is provided.

    GUI applications may prefer to create their own implementation of
    Progress which matches their application, rather
    than using this default implementation
    """

    def __init__(self):
        self.IncrementThreshold = 5
        self.HandleToWaitDialog = None
        self.DialogText = 'Please Wait'
        self.DialogTitle = ''
        self.ProgressValue = 0
        self.Hold = False
        self.show_progressBar = False

    def __del__(self):
        self.hide()

    def show_and_hold(self, text='Please wait'):
        self.hide()
        self.DialogTitle = remove_html(text)
        self.DialogText = ''
        self.ProgressValue = 0
        self.Hold = True
        self._update()

    def hide(self):
        self.DialogTitle = 'Please wait'
        self.show_progressBar = False
        if self.HandleToWaitDialog:
            self.HandleToWaitDialog.close()
            del self.HandleToWaitDialog
            self.HandleToWaitDialog = None
        self.Hold = False

    def complete(self):
        """Call to complete a progress operation, which will also hide the
        dialog unless the dialog is being held"""

        self.show_progressBar = False
        self.ProgressValue = 100
        if not self.Hold:
            self.hide()
        else:
            self._update()

    def set_progress_text(self, text='Please wait'):
        self.DialogText = remove_html(text)
        self._update()

    def set_progress_value(self, progress_value):
        self.show_progressBar = True
        self._update()

        # Only call _update() when necessary
        if abs(progress_value - self.ProgressValue) >= self.IncrementThreshold:
            self.ProgressValue = progress_value
            self._update()

    def set_progress_and_message(self, progress_value, text):
        self.show_progressBar = True
        self.DialogText = remove_html(text)
        self.ProgressValue = progress_value
        self._update()

    def cancel_clicked(self):
        return False

    def resize(self, panel_position):
        pass

    def _update(self):
        text = self.DialogText or self.DialogTitle

        if self.HandleToWaitDialog:
            self.HandleToWaitDialog.update(self.ProgressValue)
            self.HandleToWaitDialog.set_description(text)
        else:
            self.HandleToWaitDialog = tqdm(total=100, desc=text,
                                           colour='green', leave=False)


@dataclass
class ProgressStackItem:
    """Used for handling a nested progress bar

    ProgressStackItem is part of the mechanism used to nest progress
    reporting, so that for example, if an operation is performed 4 times,
    the progress bar will not go from 0% to 100% 3 times, but instead go
    from 0% to 25% for the first operation, etc."""

    progress_text: str
    min_position: int
    max_position: int
    last_progress_value: int = 0
    visible: bool = False



class CoreException(Exception):
    def __init__(self, message, identifier=None):
        super().__init__(message)
        self.identifier = id

APP_NAME = "pyreporting"
APP_AUTHOR = "CodeChoreography"


import inspect
import itertools

from pyreporting.core import CoreException


def get_calling_function(levels_to_ignore):
    """Obtain the name of the function which signalled the error (useful in
    error reporting)"""
    max_levels = 10
    full_stack = inspect.stack()
    for frame in itertools.islice(full_stack, levels_to_ignore, max_levels):
        calling_function = frame.function
        if not calling_function.startswith('CoreReporting'):
            return calling_function
    return ''


@staticmethod
def throw_exception(identifier, message, exception=None):
    if exception:
        raise CoreException(message=message, identifier=identifier) \
            from exception
    else:
        raise CoreException(message=message, identifier=identifier)
